/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "./libhw2/Utils.h"
#include "./FileReader.h"
#include "./FormReader.h"
#include "./HttpConnection.h"
#include "./HttpRequest.h"
#include "./HttpUtils.h"
#include "./HttpServer.h"

using std::cerr;
using std::cout;
using std::endl;

namespace hw3 {

// This is the function that threads are dispatched into
// in order to process new client connections.
void HttpServer_ThrFn(ThreadPool::Task *t);

// Given a request, produce a response.
HttpResponse ProcessRequest(const HttpRequest &req,
                            const std::string &basedir);

bool HttpServer::Run(void) {
  // Create the server listening socket.
  int listen_fd;
  cout << "  creating and binding the listening socket..." << endl;
  if (!ss_.BindAndListen(AF_UNSPEC, &listen_fd)) {
    cerr << endl << "Couldn't bind to the listening socket." << endl;
    return false;
  }

  // Spin, accepting connections and dispatching them.  Use a
  // threadpool to dispatch connections into their own thread.
  cout << "  accepting connections..." << endl << endl;
  ThreadPool tp(kNumThreads);
  while (1) {
    HttpServerTask *hst = new HttpServerTask(HttpServer_ThrFn);
    hst->basedir = staticfile_dirpath_;
    if (!ss_.Accept(&hst->client_fd,
                    &hst->caddr,
                    &hst->cport,
                    &hst->cdns,
                    &hst->saddr,
                    &hst->sdns)) {
      // The accept failed for some reason, so quit out of the server.
      break;
    }
    // The accept succeeded; dispatch it.
    tp.Dispatch(hst);
  }
  return true;
}

void HttpServer_ThrFn(ThreadPool::Task *t) {
  // Cast back our HttpServerTask structure with all of our new
  // client's information in it.
  HttpServerTask *hst = static_cast<HttpServerTask *>(t);
  cout << "  client " << hst->cdns << ":" << hst->cport << " "
       << "(IP address " << hst->caddr << ")" << " connected." << endl;

  // Read in the next request, process it, write the response.
  HttpConnection cn(hst->client_fd);
  HttpRequest rq;
  bool done = false;
  while ((!done) && cn.GetNextRequest(&rq)) {
    HttpResponse resp = ProcessRequest(rq, hst->basedir);
    cn.WriteResponse(resp);

    // If the client sent a "Connection: close" header, shut down
    // the connection.
    if (rq.headers["connection"] == std::string("close"))
      done = true;
  }
  delete hst;
}

HttpResponse ProcessRequest(const HttpRequest &req,
                            const std::string &basedir) {
  // Is the user asking for a static file?
  if (req.URI.substr(0, 8) == "/static/") {
    HttpResponse ret;
    // STEP 1 -- finish the implementation here.  Figure out the
    // filename the user is asking for; you probably want to make use
    // of your URLParser from HTTPUtils.cc/.h to extract the path part
    // of the URL, and then extract the substring after "static" as
    // the filename.
    //
    // Next, read that file into memory, using your FileReader class.
    // The base directory is "basedir" -- you'll need to construct the
    // appropriate filename using it.
    //
    // Finally, populate "ret" with appropriate fields, and return it.
    // The HTTP response code for a successful response is 200, and
    // the HTTP response message should be "OK".  You don't need to
    // populate any headers.
    return ret;
  }

  // Did the user post a form?
  if (req.URI.substr(0, 5) == "/post") {
    HttpResponse ret;
    FormReader f(req.URI);
    if (f.ProcessForm(&ret.body)) {
      ret.protocol = "HTTP/1.1";
      ret.response_code = 200;
      ret.message = "OK";
      return ret;
    }
    ret.protocol = "HTTP/1.1";
    ret.response_code = 200;
    ret.message = "OK";
    ret.body = "<html><body>Couldn't process form \"";
    ret.body += EscapeHTMLInPlace(req.URI);
    ret.body += "\"</body></html>";
    return ret;
  }

  // Default response for other URLs.
  HttpResponse ret;
  ret.protocol = "HTTP/1.1";
  ret.response_code = 404;
  ret.message = "Not Found";
  ret.body = "<html><body>Couldn't find \"";
  ret.body += EscapeHTMLInPlace(req.URI);
  ret.body += "\"</body></html>";
  return ret;
}

}  // namespace hw3
